#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.algo - distance de Jaccard (dictionnaires) - correction
# 
# De la distance de [Jaccard](https://fr.wikipedia.org/wiki/Indice_et_distance_de_Jaccard) à la distance de [Levenshtein](https://fr.wikipedia.org/wiki/Distance_de_Levenshtein).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Exercice 1 : Constuire l'ensemble des lettres supprimées et ajoutées
# 
# Deux mots n'ont pas forcément la même longueur, il est donc difficile de les comparer lettre à lettre. De même, on ne doit pas tenir compte de l'ordre des lettres dans chaque mot. Cette contrainte peut paraître plus difficile à mettre en oeuvre. Pourtant, si on construit un résultat intermédiaire : le décompte de chaque lettre dans un mot.

# In[2]:


def compte_lettre(mot):
    d = {}
    for c in mot:
        d[c] = d.get(c,0) + 1
    return d

compte_lettre("lettre"), compte_lettre("etre")


# Une lettre est ajoutée ou supprimée entre deux mots si son décompe est différent dans les deux dictionnaires. On s'en sert pour contruire la liste des lettres supprimées et ajoutées. Cela revient à construire la différence entre deux dictionnaires.

# In[3]:


mot1 = "lettre"
mot2 = "etres"

d1 = compte_lettre(mot1)
d2 = compte_lettre(mot2)

suppression = {}
for l in d1:
    c1 = d1[l]
    c2 = d2.get(l, 0)  # la lettre l n'appartient pas forcément au second mot
    if c2 != c1:
        suppression[l] = c2 - c1
        
ajout = {}
for l in d2:
    if l not in d1:
        c1 = 0
        c2 = d2[l]
        if c2 != c1:
            ajout[l] = c2 - c1 
    else:
        # on a déjà compté les lettres présentes dans les deux mots
        # lors de la première boucle
        pass
        
suppression, ajout        


# In[4]:





# ## Exercice 2 : écrire une fonction qui calcule la distance de Jaccard
# 
# On copie et on colle le code précédent pour créer la distance.

# In[5]:


def jaccard(mot1, mot2):    
    d1 = compte_lettre(mot1)
    d2 = compte_lettre(mot2)

    suppression = {}
    for l in d1:
        c1 = d1[l]
        c2 = d2.get(l, 0)  # la lettre l n'appartient pas forcément au second mot
        if c2 != c1:
            suppression[l] = c2 - c1

    ajout = {}
    for l in d2:
        if l not in d1:
            c1 = 0
            c2 = d2[l]
            if c2 != c1:
                ajout[l] = c2 - c1 
        else:
            # on a déjà compté les lettres présentes dans les deux mots
            # lors de la première boucle
            pass
        
    dist = sum(abs(x) for x in suppression.values()) + sum(abs(x) for x in ajout.values())
    return dist

jaccard("lettre", "etre"), jaccard("lettre", "etres")


# In[6]:


jaccard("marie", "aimer")


# In[7]:


jaccard("hôpital", "hospital")


# ## Exercice 3 :  calculer la matrice des distances
# 
# Tout d'abord, on découpe un texte en mot. On remplace les tirets en espaces.

# In[8]:


texte = "à combien sont ces six saucissons-ci, ces six saucisson-ci sont à six sous".replace("-", " ").split()
texte


# ### liste de listes

# In[9]:


distl = [ [ jaccard(texte[i], texte[j]) for i in range(len(texte))] for j in range(len(texte))]
distl


# ### dictionnaire

# In[10]:


distd = { (w,y): jaccard(w,y) for w in texte for y in texte}
distd


# On compare le nombre de coefficients :

# In[11]:


sum(len(_) for _ in distl)


# In[12]:


len(distd)


# Le dictionnaire paraît une meilleure option en terme de nombre de coefficients. Mais en terme d'espace mémoire occupé, le dictionnaire est largement plus conséquent.

# In[13]:


import sys
sys.getsizeof(distl), sys.getsizeof(distd)


# Pour ce petit exemple, la matrice paraît une bonne option mais lorsque les mots sont répétés un grand nombre de fois :

# In[14]:


texte_grand = texte * 40

distl = [ [ jaccard(texte_grand[i], texte_grand[j]) for i in range(len(texte_grand))] for j in range(len(texte_grand))]
distd = { (w,y): jaccard(w,y) for w in texte_grand for y in texte_grand}
sum(len(_) for _ in distl), len(distd)


# In[15]:


sys.getsizeof(distl), sys.getsizeof(distd)


# Le dictionnaire permet également de vérifier aisément qu'un calcul a déjà été effectué afin de gagner du temps et de ne pas le refaire.

# In[16]:


distd = {}
for w in texte_grand:
    for y in texte_grand:
        if (w,y) not in distd:
            distd[w,y] = jaccard(w,y) 
            distd[y,w] = distd[w,y]
len(distd)            


# In[17]: